/*
 * BiZduinoRTC.h - BiZduino DS1388 RTC ライブラリ
 * JeeLabs RTCLib ライブラリを元に作成
 */

#include <Wire.h>
#include "BiZduinoRTC.h"
#include <arduino.h>


// DS1388 アドレス
#define DS1388_ADDRESS          0x68
// DS1388 EEPROM レジスタ ビット
#define DS1388_EEPROM_0         0x01
#define DS1388_EEPROM_1         0x02

#define SECONDS_PER_DAY         86400L

#define SECONDS_FROM_1970_TO_2000 946684800


////////////////////////////////////////////////////////////////////////////////
// utility code, some of this could be exposed in the DateTime API if needed

static const uint8_t daysInMonth [] PROGMEM = {
  31,28,31,30,31,30,31,31,30,31,30,31
};

// number of days since 2000/01/01, valid for 2001..2099
static uint16_t date2days(uint16_t y, uint8_t m, uint8_t d) {
    if (y >= 2000)
        y -= 2000;
    uint16_t days = d;
    for (uint8_t i = 1; i < m; ++i)
        days += pgm_read_byte(daysInMonth + i - 1);
    if (m > 2 && y % 4 == 0)
        ++days;
    return days + 365 * y + (y + 3) / 4 - 1;
}

static long time2long(uint16_t days, uint8_t h, uint8_t m, uint8_t s) {
    return ((days * 24L + h) * 60 + m) * 60 + s;
}

static uint8_t conv2d(const char* p) {
    uint8_t v = 0;
    if ('0' <= *p && *p <= '9')
        v = *p - '0';
    return 10 * v + *++p - '0';
}


////////////////////////////////////////////////////////////////////////////////
// BiZduinoRTC クラス

// コンストラクタ
// t: エポックタイム（タイムゾーン込みの値）
DateTime::DateTime (uint32_t t) {
    t -= SECONDS_FROM_1970_TO_2000;    // bring to 2000 timestamp from 1970
    ss = t % 60;
    t /= 60;
    mm = t % 60;
    t /= 60;
    hh = t % 24;
    uint16_t days = t / 24;
    uint8_t leap;
    for (yOff = 0; ; ++yOff) {
        leap = yOff % 4 == 0;
        if (days < 365 + leap)
            break;
        days -= 365 + leap;
    }
    for (m = 1; ; ++m) {
        uint8_t daysPerMonth = pgm_read_byte(daysInMonth + m - 1);
        if (leap && m == 2)
            ++daysPerMonth;
        if (days < daysPerMonth)
            break;
        days -= daysPerMonth;
    }
    d = days + 1;
}

// コンストラクタ
// year: 年, month: 月, day: 日, hour: 時, min: 分, sec: 秒
DateTime::DateTime (uint16_t year, uint8_t month, uint8_t day, uint8_t hour, uint8_t min, uint8_t sec) {
    if (year >= 2000)
        year -= 2000;
    yOff = year;
    m = month;
    d = day;
    hh = hour;
    mm = min;
    ss = sec;
}

// コンストラクタ
// date: 年月日の文字列, time: 時分秒の文字列
DateTime::DateTime (const char* date, const char* time) {
    // sample input: date = "Dec 26 2009", time = "12:34:56"
    yOff = conv2d(date + 9);
    // Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec
    switch (date[0]) {
        case 'J': m = date[1] == 'a' ? 1 : m = date[2] == 'n' ? 6 : 7; break;
        case 'F': m = 2; break;
        case 'A': m = date[2] == 'r' ? 4 : 8; break;
        case 'M': m = date[2] == 'r' ? 3 : 5; break;
        case 'S': m = 9; break;
        case 'O': m = 10; break;
        case 'N': m = 11; break;
        case 'D': m = 12; break;
    }
    d = conv2d(date + 4);
    hh = conv2d(time);
    mm = conv2d(time + 3);
    ss = conv2d(time + 6);
}

// 曜日を取得
// return: 0=日, 1=月, 2=火, 3=水, 4=木, 5=金, 6=土
uint8_t DateTime::dayOfWeek() const {
    uint16_t day = get() / SECONDS_PER_DAY;
    return (day + 6) % 7; // Jan 1, 2000 is a Saturday, i.e. returns 6
}

long DateTime::get() const {
    uint16_t days = date2days(yOff, m, d);
    return time2long(days, hh, mm, ss);
}


////////////////////////////////////////////////////////////////////////////////
// BiZduinoRTC クラス

uint8_t BiZduinoRTC::WDSeconds = bin2bcd(60); //default to 60 seconds;
uint8_t BiZduinoRTC::WDTSeconds = bin2bcd(0); //default to 60.00 seconds;

// PUBLIC FUNCTIONS

// 初期化
void BiZduinoRTC::begin() {
    Wire.begin();
}

// 時刻設定(DateTime型)
// dt       設定時刻の格納先
// return:  true=成功 / false=エラー
bool BiZduinoRTC::adjust(const DateTime& dt) {
  Wire.beginTransmission(DS1388_ADDRESS);
    Wire.write((byte) 0);
    mSec = 0;
    Wire.write(bin2bcd(0));
    Wire.write(bin2bcd(dt.second()));
    Wire.write(bin2bcd(dt.minute()));
    Wire.write(bin2bcd(dt.hour()));
    Wire.write(bin2bcd(dt.dayOfWeek()));
    Wire.write(bin2bcd(dt.day()));
    Wire.write(bin2bcd(dt.month()));
    Wire.write(bin2bcd(dt.year() - 2000));
  if (Wire.endTransmission() != 0) {
    exists = false;
    return false;
  }
  exists = true;
  return true;
}

// 時刻取得(DateTime型)
// dt       時刻の格納先
// return:  true=成功 / false=エラー
bool BiZduinoRTC::read(DateTime& dt) {
  Wire.beginTransmission(DS1388_ADDRESS);
    Wire.write((byte) 0);
  if (Wire.endTransmission() != 0) {
    exists = false;
    return false;
  }
  exists = true;

  Wire.requestFrom(DS1388_ADDRESS, 8);
  if (Wire.available() < 8) return false;
  mSec = bcd2bin(Wire.read()) * 10;
  uint8_t ss = bcd2bin(Wire.read() & 0x7F);
  uint8_t mm = bcd2bin(Wire.read() & 0x7f);
  uint8_t hh = bcd2bin(Wire.read() & 0x3f);
  Wire.read();
  uint8_t d = bcd2bin(Wire.read());
  uint8_t m = bcd2bin(Wire.read() & 0x1f);
  uint16_t y = bcd2bin(Wire.read()) + 2000;
  dt = DateTime (y, m, d, hh, mm, ss);
  return true;
}


// データ有効判定
// return:  true=データ有効 / false=データ無効
//bool chipPresent()

// get()/read()実行時のミリ秒を取得する
// return:  ミリ秒
//uint16_t milliSecond()


// RTC EEPROM へ1byte書込み
// pos		バイト位置(0～511)
// c		書込み値
// return:  true=成功 / false=エラー
bool BiZduinoRTC::writeEEPROM(uint16_t pos, uint8_t c) {
  if(pos >= 512){
    return false;
  }
  int res = 0;
  uint8_t rel_pos = pos % 256;
  // Set address
  Wire.beginTransmission(getEEPROMBank(pos));
    Wire.write((byte)rel_pos);
    // Wite data
    Wire.write((byte)c);
  res = Wire.endTransmission();
  delay(10);
  return res == 0;
}

// RTC EEPROM から1byte読取り
// pos		バイト位置(0～511)
// return:  読取り値
uint8_t BiZduinoRTC::readEEPROM(uint16_t pos) {
  if(pos >= 512){
    return 0;
  }
  uint8_t rel_pos = pos % 256;
  uint8_t c = 0xFF;
  uint8_t devAddr = getEEPROMBank(pos);
  Wire.beginTransmission(devAddr);
  // Set address
  Wire.write((byte)rel_pos);
  Wire.endTransmission(true); // Stay open
  // Request one byte
  Wire.requestFrom(devAddr, (uint8_t)1);
  if (Wire.available()) {
      c = Wire.read();
  }
  return c;
}

// RTC EEPROM ページを指定して書込み
// page		ページ位置(8バイト単位のブロック、0～63)
// data		書込みデータ(8バイト)
// return:  true=成功 / false=エラー
bool BiZduinoRTC::writeEEPROMPage(uint8_t page, uint8_t *data) {
  if(page >= 64){
    return false;
  }
  int res = 0;
  Wire.beginTransmission(getEEPROMBank((uint16_t)page * 8));
  uint8_t rel_pos =((uint16_t)page * 8) % 256;
  Wire.write((byte)rel_pos);
  for(uint8_t i=0; i<8; i++){
    Wire.write((byte)data[i]);
  }
  res = Wire.endTransmission();
  delay(10);
  return res == 0;
}

// RTC EEPROM ページを指定して読取り
// page		ページ位置(8バイト単位のブロック、0～63)
// data		読取りデータ格納先(8バイト)
// return:  true=成功 / false=エラー
bool BiZduinoRTC::readEEPROMPage(uint8_t page, uint8_t *data) {
  if(page >= 64){
    return false;
  }
  int res = 0;
  uint8_t devAddr = getEEPROMBank((uint16_t)page * 8);
  Wire.beginTransmission(devAddr);
  uint8_t rel_pos =((uint16_t)page * 8) % 256;
  // Set address
  Wire.write((byte)rel_pos);
  res = Wire.endTransmission(true); // Stay open
  if (res != 0) {
      return false;
  }
  // Request 8 byte
  Wire.requestFrom(devAddr, (uint8_t)8);
  for(uint8_t i=0; i<8; i++){
    data[i] = Wire.read();
  }
  res = Wire.endTransmission();
  return res == 0;
}


// PROTECTED FUNCTIONS

uint8_t BiZduinoRTC::getEEPROMBank(uint16_t pos) {
  if(pos > 255){
    return DS1388_ADDRESS | DS1388_EEPROM_1;
  } else {
    return DS1388_ADDRESS | DS1388_EEPROM_0;
  }
}


// スタティック変数初期化
bool BiZduinoRTC::exists = false;
uint16_t BiZduinoRTC::mSec = 0;

BiZduinoRTC RTC = BiZduinoRTC(); // RTCインスタンス
