// NTPから時刻取得してRTCに設定するサンプルスケッチ
// SW3ボタンを押すとNTPサーバから時刻を取得してRTCに設定します
// Wi-Fiアクセスポイント環境に合わせて次のSSIDとパスワードを変更

#include <BiZduino.h>
#include <BiZduinoWiFi.h>
#include <BiZduinoRTC.h>

#define SSID        "<SSID>"
#define PASSWORD    "<PASSWORD>"

// NTPのサーバ名
#define NTP_HOST    "ntp.nict.jp"
// NTPのポート番号
#define NTP_PORT    123

// タイムゾーン補正値(秒)
#define TIME_ZONE 9 * 60 * 60

// NTPパケットサイズ
#define NTP_PACKET_SIZE 48

#define NTP_VERSION_4     0b00100000
#define NTP_MODE_CLIENT   0b00000011


BiZduinoWiFi wifi;  // WiFiインスタンス
BiZduinoLED3 LED3;  // LED3インスタンス
BiZduinoSW3 SW3;    // SW3インスタンス

// 曜日文字列
const char *weekName[7] = {
  "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

DateTime dt;

void setup() {
  // WiFi開始
  wifi.begin();
  RTC.begin();

  // Stationモードにする
  if (!wifi.setOprToStation()) {
    wifi.disableSerial();
    Serial.print("to station err\r\n");
  }
  // シングルコネクションモードにする
  if (!wifi.disableMUX()) {
    wifi.disableSerial();
    Serial.print("single err\r\n");
  }

  // WiFiモジュールのシリアルを切る
  wifi.disableSerial();

}

void loop() {
  static unsigned long lastMilli = 0;
  static uint8_t mux_id = 0;
  uint8_t buffer[256] = {0};

  // SW3が押されているか判定
  if (SW3.isClicked()) {

    // LED3点灯
    LED3.on();

    // APに接続する
    if (wifi.joinAP(SSID, PASSWORD)) {

      // NTPからエポックタイムを取得
      unsigned long epoch = getNTP();

      // タイムゾーンを足してRTCに設定
      dt = DateTime(epoch + TIME_ZONE);
      if (RTC.adjust(dt)) {
        wifi.disableSerial();
        Serial.println("RTC set success!");
      } else {
        wifi.disableSerial();
        Serial.println("RTC set error!");
      }

      // APを切断
      wifi.leaveAP();

    } else {
      wifi.disableSerial();
      Serial.print("Join AP failure\r\n");
    }

    // LED3消灯
    LED3.off();

    // WiFiモジュールのシリアルを切る
    wifi.disableSerial();

  } else {

    unsigned long nowMilli = millis();
    if ((nowMilli - lastMilli) >= 1000) {  // 経過時間

      // RTCの時刻表示
      printRTC();

      lastMilli = nowMilli;
    }

  }
}

// NTPから時刻を取得する
unsigned long getNTP()
{
  unsigned long epoch = 0;

  // NTP パケットバッファ
  byte packetBuffer[NTP_PACKET_SIZE];

  if (wifi.registerUDP(NTP_HOST, NTP_PORT)) {
    memset(packetBuffer, 0, NTP_PACKET_SIZE);
    packetBuffer[0] = NTP_VERSION_4 | NTP_MODE_CLIENT;  // LI, Version, Mode
    wifi.send(packetBuffer, NTP_PACKET_SIZE);

    uint32_t len = wifi.recv(packetBuffer, NTP_PACKET_SIZE, 10000);

    if (len == NTP_PACKET_SIZE) {
      unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
      unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
      unsigned long secsSince1900 = highWord << 16 | lowWord;
      const unsigned long seventyYears = 2208988800UL;
      epoch = secsSince1900 - seventyYears;   // UNIXタイムスタンプ
    }

    wifi.unregisterUDP();

  } else {
    wifi.disableSerial();
    Serial.println("register udp err");
  }

  return epoch;
}


// RTCの時刻表示
void printRTC() {

  // RTCの時刻を取得
  if (RTC.read(dt)) {

    Serial.print(dt.year());    // 年の表示
    Serial.write('/');
    print2digits(dt.month());   // 月の表示
    Serial.write('/');
    print2digits(dt.day());     // 日の表示
    Serial.write(' ');
    Serial.print(weekName[dt.dayOfWeek()]); // 曜日の表示
    Serial.write(' ');
    print2digits(dt.hour());    // 時の表示
    Serial.write(':');
    print2digits(dt.minute());  // 分の表示
    Serial.write(':');
    print2digits(dt.second());  // 秒の表示
    Serial.println();
  } else {
    if (RTC.chipPresent()) {
      // RTCデータエラー
      Serial.println("RTC is stopped.");
      Serial.println();
    } else {
      // RTC通信エラー
      Serial.println("RTC read error!");
      Serial.println();
    }
  }

}


// 2桁表示
void print2digits(int number) {
  if (number >= 0 && number < 10) {
    Serial.write('0');
  }
  Serial.print(number);
}

